# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import dataclass

from mautrix.types import EventID, RoomID
from mautrix.util.async_db import Database

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class Message:
    db: ClassVar[Database] = fake_db

    mxid: EventID
    mx_room: RoomID
    item_id: str
    client_context: str | None
    receiver: int
    sender: int
    ig_timestamp: int | None

    @property
    def ig_timestamp_ms(self) -> int:
        return (self.ig_timestamp // 1000) if self.ig_timestamp else 0

    async def insert(self) -> None:
        q = """
            INSERT INTO message (mxid, mx_room, item_id, client_context, receiver, sender,
                                 ig_timestamp)
            VALUES ($1, $2, $3, $4, $5, $6, $7)
        """
        await self.db.execute(
            q,
            self.mxid,
            self.mx_room,
            self.item_id,
            self.client_context,
            self.receiver,
            self.sender,
            self.ig_timestamp,
        )

    async def delete(self) -> None:
        q = "DELETE FROM message WHERE item_id=$1 AND receiver=$2"
        await self.db.execute(q, self.item_id, self.receiver)

    @classmethod
    async def delete_all(cls, room_id: RoomID) -> None:
        await cls.db.execute("DELETE FROM message WHERE mx_room=$1", room_id)

    _columns = "mxid, mx_room, item_id, client_context, receiver, sender, ig_timestamp"

    @classmethod
    async def get_by_mxid(cls, mxid: EventID, mx_room: RoomID) -> Message | None:
        q = f"SELECT {cls._columns} FROM message WHERE mxid=$1 AND mx_room=$2"
        row = await cls.db.fetchrow(q, mxid, mx_room)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def get_last(cls, mx_room: RoomID) -> Message | None:
        q = f"""
        SELECT {cls._columns} FROM message
        WHERE mx_room=$1 AND ig_timestamp IS NOT NULL AND item_id NOT LIKE 'fi.mau.instagram.%'
        ORDER BY ig_timestamp DESC LIMIT 1
        """
        row = await cls.db.fetchrow(q, mx_room)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def get_closest(cls, mx_room: RoomID, before_ts: int) -> Message | None:
        q = f"""
        SELECT {cls._columns} FROM message
        WHERE mx_room=$1 AND ig_timestamp<=$2 AND item_id NOT LIKE 'fi.mau.instagram.%'
        ORDER BY ig_timestamp DESC LIMIT 1
        """
        row = await cls.db.fetchrow(q, mx_room, before_ts)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def get_by_item_id(cls, item_id: str, receiver: int) -> Message | None:
        q = f"SELECT {cls._columns} FROM message WHERE item_id=$1 AND receiver=$2"
        row = await cls.db.fetchrow(q, item_id, receiver)
        if not row:
            return None
        return cls(**row)

    @property
    def is_internal(self) -> bool:
        return self.item_id.startswith("fi.mau.instagram.")



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Type, TypeVar
import json

from ..types import CurrentUserResponse
from .base import BaseAndroidAPI

T = TypeVar("T")


class AccountAPI(BaseAndroidAPI):
    async def current_user(self) -> CurrentUserResponse:
        return await self.std_http_get(
            f"/api/v1/accounts/current_user/",
            query={"edit": "true"},
            response_type=CurrentUserResponse,
        )

    async def set_biography(self, text: str) -> CurrentUserResponse:
        # TODO entities?
        return await self.__command("set_biography", device_id=self.state.device.id, raw_text=text)

    async def set_profile_picture(self, upload_id: str) -> CurrentUserResponse:
        return await self.__command(
            "change_profile_picture", use_fbuploader="true", upload_id=upload_id
        )

    async def remove_profile_picture(self) -> CurrentUserResponse:
        return await self.__command("remove_profile_picture")

    async def set_private(self, private: bool) -> CurrentUserResponse:
        return await self.__command("set_private" if private else "set_public")

    async def confirm_email(self, slug: str) -> CurrentUserResponse:
        # slug can contain slashes, but it shouldn't start or end with one
        return await self.__command(f"confirm_email/{slug}")

    async def send_recovery_flow_email(self, query: str):
        req = {
            "_csrftoken": self.state.cookies.csrf_token,
            "adid": "",
            "guid": self.state.device.uuid,
            "device_id": self.state.device.id,
            "query": query,
        }
        # TODO parse response content
        return await self.std_http_post(f"/api/v1/accounts/send_recovery_flow_email/", data=req)

    async def edit_profile(
        self,
        external_url: str | None = None,
        gender: str | None = None,
        phone_number: str | None = None,
        username: str | None = None,
        # TODO should there be a last_name?
        first_name: str | None = None,
        biography: str | None = None,
        email: str | None = None,
    ) -> CurrentUserResponse:
        return await self.__command(
            "edit_profile",
            device_id=self.state.device.id,
            email=email,
            external_url=external_url,
            first_name=first_name,
            username=username,
            phone_number=phone_number,
            gender=gender,
            biography=biography,
        )

    async def __command(
        self, command: str, response_type: Type[T] = CurrentUserResponse, **kwargs: str
    ) -> T:
        req = {
            "_csrftoken": self.state.cookies.csrf_token,
            "_uid": self.state.cookies.user_id,
            "_uuid": self.state.device.uuid,
            **kwargs,
        }
        return await self.std_http_post(
            f"/api/v1/accounts/{command}/",
            data=req,
            filter_nulls=True,
            response_type=response_type,
        )

    async def read_msisdn_header(self, usage: str = "default"):
        req = {
            "mobile_subno_usage": usage,
            "device_id": self.state.device.uuid,
        }
        return await self.std_http_post("/api/v1/accounts/read_msisdn_header/", data=req)

    async def msisdn_header_bootstrap(self, usage: str = "default"):
        req = {
            "mobile_subno_usage": usage,
            "device_id": self.state.device.uuid,
        }
        return await self.std_http_post("/api/v1/accounts/msisdn_header_bootstrap/", data=req)

    async def contact_point_prefill(self, usage: str = "default"):
        req = {
            "mobile_subno_usage": usage,
            "device_id": self.state.device.uuid,
        }
        return await self.std_http_post("/api/v1/accounts/contact_point_prefill/", data=req)

    async def get_prefill_candidates(self):
        req = {
            "android_device_id": self.state.device.id,
            "usages": json.dumps(["account_recovery_omnibox"]),
            "device_id": self.state.device.uuid,
        }
        # TODO parse response content
        return await self.std_http_post("/api/v1/accounts/get_prefill_candidates/", data=req)

    async def process_contact_point_signals(self):
        req = {
            "phone_id": self.state.device.phone_id,
            "_csrftoken": self.state.cookies.csrf_token,
            "_uid": self.state.cookies.user_id,
            "device_id": self.state.device.uuid,
            "_uuid": self.state.device.uuid,
            "google_tokens": json.dumps([]),
        }
        return await self.std_http_post(
            "/api/v1/accounts/process_contact_point_signals/", data=req
        )



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import dataclass
import asyncpg

from mautrix.types import ContentURI, RoomID, UserID
from mautrix.util.async_db import Database

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class Portal:
    db: ClassVar[Database] = fake_db

    thread_id: str
    receiver: int
    other_user_pk: int | None
    mxid: RoomID | None
    name: str | None
    avatar_url: ContentURI | None
    encrypted: bool
    name_set: bool
    avatar_set: bool
    relay_user_id: UserID | None

    @property
    def _values(self):
        return (
            self.thread_id,
            self.receiver,
            self.other_user_pk,
            self.mxid,
            self.name,
            self.avatar_url,
            self.encrypted,
            self.name_set,
            self.avatar_set,
            self.relay_user_id,
        )

    async def insert(self) -> None:
        q = (
            "INSERT INTO portal (thread_id, receiver, other_user_pk, mxid, name, avatar_url, "
            "                    encrypted, name_set, avatar_set, relay_user_id) "
            "VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)"
        )
        await self.db.execute(q, *self._values)

    async def update(self) -> None:
        q = (
            "UPDATE portal SET other_user_pk=$3, mxid=$4, name=$5, avatar_url=$6, encrypted=$7,"
            "                  name_set=$8, avatar_set=$9, relay_user_id=$10 "
            "WHERE thread_id=$1 AND receiver=$2"
        )
        await self.db.execute(q, *self._values)

    @classmethod
    def _from_row(cls, row: asyncpg.Record) -> Portal:
        return cls(**row)

    @classmethod
    async def get_by_mxid(cls, mxid: RoomID) -> Portal | None:
        q = (
            "SELECT thread_id, receiver, other_user_pk, mxid, name, avatar_url, encrypted, "
            "       name_set, avatar_set, relay_user_id "
            "FROM portal WHERE mxid=$1"
        )
        row = await cls.db.fetchrow(q, mxid)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def get_by_thread_id(
        cls, thread_id: str, receiver: int, rec_must_match: bool = True
    ) -> Portal | None:
        q = (
            "SELECT thread_id, receiver, other_user_pk, mxid, name, avatar_url, encrypted, "
            "       name_set, avatar_set, relay_user_id "
            "FROM portal WHERE thread_id=$1 AND receiver=$2"
        )
        if not rec_must_match:
            q = (
                "SELECT thread_id, receiver, other_user_pk, mxid, name, avatar_url, encrypted, "
                "       name_set, avatar_set "
                "FROM portal WHERE thread_id=$1 AND (receiver=$2 OR receiver=0)"
            )
        row = await cls.db.fetchrow(q, thread_id, receiver)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def find_private_chats_of(cls, receiver: int) -> list[Portal]:
        q = (
            "SELECT thread_id, receiver, other_user_pk, mxid, name, avatar_url, encrypted, "
            "       name_set, avatar_set, relay_user_id "
            "FROM portal WHERE receiver=$1 AND other_user_pk IS NOT NULL"
        )
        rows = await cls.db.fetch(q, receiver)
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def find_private_chats_with(cls, other_user: int) -> list[Portal]:
        q = (
            "SELECT thread_id, receiver, other_user_pk, mxid, name, avatar_url, encrypted, "
            "       name_set, avatar_set, relay_user_id "
            "FROM portal WHERE other_user_pk=$1"
        )
        rows = await cls.db.fetch(q, other_user)
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def find_private_chat_id(cls, receiver: int, other_user: int) -> str | None:
        q = "SELECT thread_id FROM portal WHERE receiver=$1 AND other_user_pk=$2"
        return await cls.db.fetchval(q, receiver, other_user)

    @classmethod
    async def all_with_room(cls) -> list[Portal]:
        q = (
            "SELECT thread_id, receiver, other_user_pk, mxid, name, avatar_url, encrypted, "
            "       name_set, avatar_set, relay_user_id "
            "FROM portal WHERE mxid IS NOT NULL"
        )
        rows = await cls.db.fetch(q)
        return [cls._from_row(row) for row in rows]



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import json
import pkgutil

from attr import dataclass

from mautrix.types import SerializableAttrs

default_capabilities = json.loads(
    pkgutil.get_data("mauigpapi.state", "samples/supported-capabilities.json")
)


@dataclass
class AndroidApplication(SerializableAttrs):
    # TODO these newer versions make the iris subscribe stop working for some reason
    # APP_VERSION: str = "168.0.0.40.355"
    # APP_VERSION_CODE: str = "261079771"
    # APP_VERSION: str = "167.1.0.25.120"
    # APP_VERSION_CODE: str = "259829117"
    APP_VERSION: str = "159.0.0.29.122"
    APP_VERSION_CODE: str = "244390482"
    BREADCRUMB_KEY: str = "iN4$aGr0m"
    FACEBOOK_ANALYTICS_APPLICATION_ID: str = "567067343352427"
    FACEBOOK_OTA_FIELDS: str = (
        "update%7Bdownload_uri%2Cdownload_uri_delta_base%2Cversion_code_delta_base"
        "%2Cdownload_uri_delta%2Cfallback_to_full_update%2Cfile_size_delta%2Cversion_code"
        "%2Cpublished_date%2Cfile_size%2Cota_bundle_type%2Cresources_checksum"
        "%2Cresources_sha256_checksum%2Callowed_networks%2Crelease_id%7D "
    )
    FACEBOOK_ORCA_APPLICATION_ID: str = "124024574287414"

    BLOKS_VERSION_ID: str = "0e00f30ed0184b9c914a8baad3fe538aa36a9f0faad173486e76af5ee9310d0b"
    CAPABILITIES: str = "3brTvx8="
    SUPPORTED_CAPABILITIES: str = default_capabilities

    EXPERIMENTS: str = ",".join(
        (
            "ig_android_ad_stories_scroll_perf_universe",
            "ig_android_ads_bottom_sheet_report_flow",
            "ig_android_ads_data_preferences_universe",
            "ig_android_ads_rendering_logging",
            "ig_android_canvas_cookie_universe",
            "ig_android_feed_ads_ppr_universe",
            "ig_android_graphql_survey_new_proxy_universe",
            "ig_android_viewpoint_occlusion",
            "ig_android_viewpoint_stories_public_testing",
            "ig_promote_interactive_poll_sticker_igid_universe",
            "ig_stories_ads_delivery_rules",
            "ig_stories_ads_media_based_insertion",
            "ig_android_li_session_chaining",
            "ig_android_logging_metric_universe_v2",
            "mi_viewpoint_viewability_universe",
            "ig_android_branded_content_appeal_states",
            "ig_android_branded_content_tag_redesign_organic",
            "ig_branded_content_settings_unsaved_changes_dialog",
            "ig_branded_content_tagging_approval_request_flow_brand_side_v2",
            "ig_rn_branded_content_settings_approval_on_select_save",
            "aymt_instagram_promote_flow_abandonment_ig_universe",
            "ig_android_account_insights_shopping_content_universe",
            "ig_android_business_attribute_sync",
            "ig_android_business_promote_tooltip",
            "ig_android_business_transaction_in_stories_consumer",
            "ig_android_business_transaction_in_stories_creator",
            "ig_android_create_page_on_top_universe",
            "ig_android_fb_sync_options_universe",
            "ig_android_fb_url_universe",
            "ig_android_fbpage_on_profile_side_tray",
            "ig_android_insights_post_dismiss_button",
            "ig_android_location_integrity_universe",
            "ig_android_personal_user_xposting_destination_fix",
            "ig_android_place_signature_universe",
            "ig_android_product_breakdown_post_insights",
            "ig_android_secondary_inbox_universe",
            "ig_android_share_publish_page_universe",
            "ig_android_skip_button_content_on_connect_fb_universe",
            "ig_business_new_value_prop_universe",
            "ig_android_business_cross_post_with_biz_id_infra",
            "ig_android_claim_location_page",
            "ig_android_edit_location_page_info",
            "ig_android_page_claim_deeplink_qe",
            "ig_biz_growth_insights_universe",
            "android_cameracore_fbaudio_integration_ig_universe",
            "android_ig_cameracore_aspect_ratio_fix",
            "ig_android_arengine_remote_scripting_universe",
            "ig_android_camera_formats_ranking_universe",
            "ig_android_camera_gyro_universe",
            "ig_android_camera_reduce_file_exif_reads",
            "ig_android_enable_automated_instruction_text_ar",
            "ig_android_image_exif_metadata_ar_effect_id_universe",
            "ig_android_optic_face_detection",
            "ig_android_optic_new_architecture",
            "ig_android_optic_photo_cropping_fixes",
            "ig_android_recognition_tracking_thread_prority_universe",
            "ig_android_ttcp_improvements",
            "ig_camera_android_bg_processor",
            "ig_camera_android_effect_metadata_cache_refresh_universe",
            "ig_camera_android_facetracker_v12_universe",
            "ig_camera_android_feed_effect_attribution_universe",
            "ig_camera_android_gallery_search_universe",
            "ig_camera_android_gyro_senser_sampling_period_universe",
            "ig_camera_android_paris_filter_universe",
            "ig_camera_android_share_effect_link_universe",
            "ig_camera_android_subtle_filter_universe",
            "ig_cameracore_android_new_optic_camera2",
            "ig_cameracore_android_new_optic_camera2_galaxy",
            "ig_android_external_gallery_import_affordance",
            "ig_android_feed_auto_share_to_facebook_dialog",
            "ig_android_fs_creation_flow_tweaks",
            "ig_android_fs_new_gallery",
            "ig_android_fs_new_gallery_hashtag_prompts",
            "ig_android_music_story_fb_crosspost_universe",
            "ig_android_partial_share_sheet",
            "ig_android_photo_creation_large_width",
            "ig_android_render_thread_memory_leak_holdout",
            "ig_android_direct_add_member_dialog_universe",
            "ig_android_direct_aggregated_media_and_reshares",
            "ig_android_direct_block_from_group_message_requests",
            "ig_android_direct_inbox_cache_universe",
            "ig_android_direct_leave_from_group_message_requests",
            "ig_android_direct_mark_as_read_notif_action",
            "ig_android_direct_message_follow_button",
            "ig_android_direct_multi_upload_universe",
            "ig_android_direct_mutation_manager_media_3",
            "ig_android_direct_new_gallery",
            "ig_android_direct_segmented_video",
            "ig_android_direct_state_observer",
            "ig_android_direct_thread_target_queue_universe",
            "ig_android_direct_view_more_qe",
            "ig_android_direct_wellbeing_message_reachability_settings",
            "ig_android_disable_manual_retries",
            "ig_android_gallery_grid_controller_folder_cache",
            "ig_android_multi_thread_sends",
            "ig_android_on_notification_cleared_async_universe",
            "ig_android_push_reliability_universe",
            "ig_android_wait_for_app_initialization_on_push_action_universe",
            "ig_direct_android_bubble_system",
            "ig_direct_max_participants",
            "ig_android_explore_recyclerview_universe",
            "ig_android_explore_reel_loading_state",
            "ig_android_not_interested_secondary_options",
            "ig_android_save_to_collections_bottom_sheet_refactor",
            "ig_android_save_to_collections_flow",
            "ig_android_smplt_universe",
            "ig_explore_2018_post_chaining_account_recs_dedupe_universe",
            "ig_explore_2019_h1_destination_cover",
            "ig_explore_2019_h1_video_autoplay_resume",
            "ig_explore_reel_ring_universe",
            "ig_android_feed_cache_update",
            "ig_android_view_info_universe",
            "ig_carousel_bumped_organic_impression_client_universe",
            "ig_end_of_feed_universe",
            "ig_android_igtv_autoplay_on_prepare",
            "ig_android_igtv_browse_long_press",
            "ig_android_igtv_explore2x2_viewer",
            "ig_android_igtv_pip",
            "ig_android_igtv_player_follow_button",
            "ig_android_igtv_refresh_tv_guide_interval",
            "ig_android_igtv_stories_preview",
            "ig_android_igtv_whitelisted_for_web",
            "ig_android_biz_story_to_fb_page_improvement",
            "ig_android_contact_point_upload_rate_limit_killswitch",
            "ig_android_country_code_fix_universe",
            "ig_android_dual_destination_quality_improvement",
            "ig_android_explore_discover_people_entry_point_universe",
            "ig_android_fb_follow_server_linkage_universe",
            "ig_android_fb_link_ui_polish_universe",
            "ig_android_fb_profile_integration_universe",
            "ig_android_fbc_upsell_on_dp_first_load",
            "ig_android_ig_personal_account_to_fb_page_linkage_backfill",
            "ig_android_inline_editing_local_prefill",
            "ig_android_interest_follows_universe",
            "ig_android_invite_list_button_redesign_universe",
            "ig_android_login_onetap_upsell_universe",
            "ig_android_new_follower_removal_universe",
            "ig_android_persistent_nux",
            "ig_android_qp_kill_switch",
            "ig_android_recommend_accounts_destination_routing_fix",
            "ig_android_self_following_v2_universe",
            "ig_android_self_story_button_non_fbc_accounts",
            "ig_android_self_story_setting_option_in_menu",
            "ig_android_separate_empty_feed_su_universe",
            "ig_android_show_create_content_pages_universe",
            "ig_android_show_self_followers_after_becoming_private_universe",
            "ig_android_suggested_users_background",
            "ig_android_test_not_signing_address_book_unlink_endpoint",
            "ig_android_test_remove_button_main_cta_self_followers_universe",
            "ig_android_unfollow_from_main_feed_v2",
            "ig_android_unfollow_reciprocal_universe",
            "ig_android_unify_graph_management_actions",
            "ig_android_whats_app_contact_invite_universe",
            "ig_android_xposting_feed_to_stories_reshares_universe",
            "ig_android_xposting_newly_fbc_people",
            "ig_android_xposting_reel_memory_share_universe",
            "ig_android_zero_rating_carrier_signal",
            "ig_fb_graph_differentiation",
            "ig_graph_evolution_holdout_universe",
            "ig_graph_management_h2_2019_universe",
            "ig_graph_management_production_h2_2019_holdout_universe",
            "ig_inventory_connections",
            "ig_pacing_overriding_universe",
            "ig_sim_api_analytics_reporting",
            "ig_xposting_biz_feed_to_story_reshare",
            "ig_xposting_mention_reshare_stories",
            "instagram_ns_qp_prefetch_universe",
            "ig_android_analytics_background_uploader_schedule",
            "ig_android_appstate_logger",
            "ig_android_apr_lazy_build_request_infra",
            "ig_android_camera_leak",
            "ig_android_crash_fix_detach_from_gl_context",
            "ig_android_dead_code_detection",
            "ig_android_disk_usage_logging_universe",
            "ig_android_dropframe_manager",
            "ig_android_image_upload_quality_universe",
            "ig_android_mainfeed_generate_prefetch_background",
            "ig_android_media_remodel",
            "ig_android_memory_use_logging_universe",
            "ig_android_network_perf_qpl_ppr",
            "ig_android_qpl_class_marker",
            "ig_android_react_native_email_sms_settings_universe",
            "ig_android_sharedpreferences_qpl_logging",
            "ig_disable_fsync_universe",
            "ig_mprotect_code_universe",
            "ig_prefetch_scheduler_backtest",
            "ig_traffic_routing_universe",
            "ig_android_qr_code_nametag",
            "ig_android_qr_code_scanner",
            "ig_android_live_egl10_compat",
            "ig_android_live_realtime_comments_universe",
            "ig_android_live_subscribe_user_level_universe",
            "ig_android_tango_cpu_overuse_universe",
            "igqe_pending_tagged_posts",
            "ig_android_enable_zero_rating",
            "ig_android_sso_kototoro_app_universe",
            "ig_android_sso_use_trustedapp_universe",
            "ig_android_whitehat_options_universe",
            "ig_android_payments_growth_promote_payments_in_payments",
            "ig_android_purx_native_checkout_universe",
            "ig_android_shopping_pdp_post_purchase_sharing",
            "ig_payment_checkout_cvv",
            "ig_payment_checkout_info",
            "ig_payments_billing_address",
            "ig_android_profile_thumbnail_impression",
            "ig_android_user_url_deeplink_fbpage_endpoint",
            "instagram_android_profile_follow_cta_context_feed",
            "ig_android_mqtt_cookie_auth_memcache_universe",
            "ig_android_realtime_mqtt_logging",
            "ig_rti_inapp_notifications_universe",
            "ig_android_live_webrtc_livewith_params",
            "saved_collections_cache_universe",
            "ig_android_search_nearby_places_universe",
            "ig_search_hashtag_content_advisory_remove_snooze",
            "ig_android_shopping_bag_null_state_v1",
            "ig_android_shopping_bag_optimization_universe",
            "ig_android_shopping_checkout_signaling",
            "ig_android_shopping_product_metadata_on_product_tiles_universe",
            "ig_android_wishlist_reconsideration_universe",
            "ig_biz_post_approval_nux_universe",
            "ig_commerce_platform_ptx_bloks_universe",
            "ig_shopping_bag_universe",
            "ig_shopping_checkout_improvements_universe",
            "ig_shopping_checkout_improvements_v2_universe",
            "ig_shopping_checkout_mvp_experiment",
            "ig_shopping_insights_wc_copy_update_android",
            "ig_shopping_size_selector_redesign",
            "instagram_shopping_hero_carousel_visual_variant_consolidation",
            "ig_android_audience_control",
            "ig_android_camera_upsell_dialog",
            "ig_android_create_mode_memories_see_all",
            "ig_android_create_mode_tap_to_cycle",
            "ig_android_create_mode_templates",
            "ig_android_feed_post_sticker",
            "ig_android_frx_creation_question_responses_reporting",
            "ig_android_frx_highlight_cover_reporting_qe",
            "ig_android_music_browser_redesign",
            "ig_android_publisher_stories_migration",
            "ig_android_rainbow_hashtags",
            "ig_android_recipient_picker",
            "ig_android_reel_tray_item_impression_logging_viewpoint",
            "ig_android_save_all",
            "ig_android_stories_blacklist",
            "ig_android_stories_boomerang_v2_universe",
            "ig_android_stories_context_sheets_universe",
            "ig_android_stories_gallery_sticker_universe",
            "ig_android_stories_gallery_video_segmentation",
            "ig_android_stories_layout_universe",
            "ig_android_stories_music_awareness_universe",
            "ig_android_stories_music_lyrics",
            "ig_android_stories_music_overlay",
            "ig_android_stories_music_search_typeahead",
            "ig_android_stories_project_eclipse",
            "ig_android_stories_question_sticker_music_format",
            "ig_android_stories_quick_react_gif_universe",
            "ig_android_stories_share_extension_video_segmentation",
            "ig_android_stories_sundial_creation_universe",
            "ig_android_stories_video_prefetch_kb",
            "ig_android_stories_vpvd_container_module_fix",
            "ig_android_stories_weblink_creation",
            "ig_android_story_bottom_sheet_clips_single_audio_mas",
            "ig_android_story_bottom_sheet_music_mas",
            "ig_android_story_bottom_sheet_top_clips_mas",
            "ig_android_xposting_dual_destination_shortcut_fix",
            "ig_stories_allow_camera_actions_while_recording",
            "ig_stories_rainbow_ring",
            "ig_threads_clear_notifications_on_has_seen",
            "ig_threads_sanity_check_thread_viewkeys",
            "ig_android_action_sheet_migration_universe",
            "ig_android_emoji_util_universe_3",
            "ig_android_recyclerview_binder_group_enabled_universe",
            "ig_emoji_render_counter_logging_universe",
            "ig_android_delete_ssim_compare_img_soon",
            "ig_android_reel_raven_video_segmented_upload_universe",
            "ig_android_render_output_surface_timeout_universe",
            "ig_android_sidecar_segmented_streaming_universe",
            "ig_android_stories_video_seeking_audio_bug_fix",
            "ig_android_video_abr_universe",
            "ig_android_video_call_finish_universe",
            "ig_android_video_exoplayer_2",
            "ig_android_video_fit_scale_type_igtv",
            "ig_android_video_product_specific_abr",
            "ig_android_video_qp_logger_universe",
            "ig_android_video_raven_bitrate_ladder_universe",
            "ig_android_video_raven_passthrough",
            "ig_android_video_raven_streaming_upload_universe",
            "ig_android_video_streaming_upload_universe",
            "ig_android_video_upload_hevc_encoding_universe",
            "ig_android_video_upload_quality_qe1",
            "ig_android_video_upload_transform_matrix_fix_universe",
            "ig_android_video_visual_quality_score_based_abr",
            "ig_video_experimental_encoding_consumption_universe",
            "ig_android_vc_background_call_toast_universe",
            "ig_android_vc_capture_universe",
            "ig_android_vc_codec_settings",
            "ig_android_vc_cowatch_config_universe",
            "ig_android_vc_cowatch_media_share_universe",
            "ig_android_vc_cowatch_universe",
            "ig_android_vc_cpu_overuse_universe",
            "ig_android_vc_explicit_intent_for_notification",
            "ig_android_vc_face_effects_universe",
            "ig_android_vc_join_timeout_universe",
            "ig_android_vc_migrate_to_bluetooth_v2_universe",
            "ig_android_vc_missed_call_call_back_action_universe",
            "ig_android_vc_shareable_moments_universe",
            "ig_android_comment_warning_non_english_universe",
            "ig_android_feed_post_warning_universe",
            "ig_android_image_pdq_calculation",
            "ig_android_logged_in_delta_migration",
            "ig_android_wellbeing_support_frx_cowatch_reporting",
            "ig_android_wellbeing_support_frx_hashtags_reporting",
            "ig_android_wellbeing_timeinapp_v1_universe",
            "ig_challenge_general_v2",
            "ig_ei_option_setting_universe",
        )
    )
    LOGIN_EXPERIMENTS: str = "".join(
        (
            "ig_android_device_detection_info_upload",
            "ig_android_device_info_foreground_reporting",
            "ig_android_suma_landing_page",
            "ig_android_device_based_country_verification",
            "ig_android_direct_add_direct_to_android_native_photo_share_sheet",
            "ig_android_direct_main_tab_universe_v2",
            "ig_account_identity_logged_out_signals_global_holdout_universe",
            "ig_android_fb_account_linking_sampling_freq_universe",
            "ig_android_login_identifier_fuzzy_match",
            "ig_android_nux_add_email_device",
            "ig_android_passwordless_account_password_creation_universe",
            "ig_android_reg_modularization_universe",
            "ig_android_retry_create_account_universe",
            "ig_growth_android_profile_pic_prefill_with_fb_pic_2",
            "ig_android_quickcapture_keep_screen_on",
            "ig_android_gmail_oauth_in_reg",
            "ig_android_reg_nux_headers_cleanup_universe",
            "ig_android_smartlock_hints_universe",
            "ig_android_security_intent_switchoff",
            "ig_android_sim_info_upload",
            "ig_android_caption_typeahead_fix_on_o_universe",
            "ig_android_video_render_codec_low_memory_gc",
            "ig_android_device_verification_fb_signup",
            "ig_android_device_verification_separate_endpoint",
        )
    )



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import dataclass
import asyncpg

from mauigpapi.state import AndroidState
from mautrix.types import RoomID, UserID
from mautrix.util.async_db import Database

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class User:
    db: ClassVar[Database] = fake_db

    mxid: UserID
    igpk: int | None
    state: AndroidState | None
    notice_room: RoomID | None
    seq_id: int | None
    snapshot_at_ms: int | None

    @property
    def _values(self):
        return (
            self.mxid,
            self.igpk,
            self.state.json() if self.state else None,
            self.notice_room,
            self.seq_id,
            self.snapshot_at_ms,
        )

    async def insert(self) -> None:
        q = """
        INSERT INTO "user" (mxid, igpk, state, notice_room, seq_id, snapshot_at_ms)
        VALUES ($1, $2, $3, $4, $5, $6)
        """
        await self.db.execute(q, *self._values)

    async def update(self) -> None:
        q = """
        UPDATE "user" SET igpk=$2, state=$3, notice_room=$4, seq_id=$5, snapshot_at_ms=$6
        WHERE mxid=$1
        """
        await self.db.execute(q, *self._values)

    async def save_seq_id(self) -> None:
        q = 'UPDATE "user" SET seq_id=$2, snapshot_at_ms=$3 WHERE mxid=$1'
        await self.db.execute(q, self.mxid, self.seq_id, self.snapshot_at_ms)

    @classmethod
    def _from_row(cls, row: asyncpg.Record) -> User:
        data = {**row}
        state_str = data.pop("state")
        return cls(state=AndroidState.parse_json(state_str) if state_str else None, **data)

    _columns = "mxid, igpk, state, notice_room, seq_id, snapshot_at_ms"

    @classmethod
    async def get_by_mxid(cls, mxid: UserID) -> User | None:
        q = f'SELECT {cls._columns} FROM "user" WHERE mxid=$1'
        row = await cls.db.fetchrow(q, mxid)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def get_by_igpk(cls, igpk: int) -> User | None:
        q = f'SELECT {cls._columns} FROM "user" WHERE igpk=$1'
        row = await cls.db.fetchrow(q, igpk)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def all_logged_in(cls) -> list[User]:
        q = f'SELECT {cls._columns} FROM "user" WHERE igpk IS NOT NULL'
        rows = await cls.db.fetch(q)
        return [cls._from_row(row) for row in rows]



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, AsyncIterable, Awaitable, cast
import os.path

from yarl import URL

from mauigpapi.types import BaseResponseUser
from mautrix.appservice import IntentAPI
from mautrix.bridge import BasePuppet, async_getter_lock
from mautrix.types import ContentURI, RoomID, SyncToken, UserID
from mautrix.util.simple_template import SimpleTemplate

from . import portal as p, user as u
from .config import Config
from .db import Puppet as DBPuppet

if TYPE_CHECKING:
    from .__main__ import InstagramBridge


class Puppet(DBPuppet, BasePuppet):
    by_pk: dict[int, Puppet] = {}
    by_custom_mxid: dict[UserID, Puppet] = {}
    hs_domain: str
    mxid_template: SimpleTemplate[int]

    config: Config

    default_mxid_intent: IntentAPI
    default_mxid: UserID

    def __init__(
        self,
        pk: int,
        name: str | None = None,
        username: str | None = None,
        photo_id: str | None = None,
        photo_mxc: ContentURI | None = None,
        name_set: bool = False,
        avatar_set: bool = False,
        is_registered: bool = False,
        custom_mxid: UserID | None = None,
        access_token: str | None = None,
        next_batch: SyncToken | None = None,
        base_url: URL | None = None,
    ) -> None:
        super().__init__(
            pk=pk,
            name=name,
            username=username,
            photo_id=photo_id,
            name_set=name_set,
            photo_mxc=photo_mxc,
            avatar_set=avatar_set,
            is_registered=is_registered,
            custom_mxid=custom_mxid,
            access_token=access_token,
            next_batch=next_batch,
            base_url=base_url,
        )
        self.log = self.log.getChild(str(pk))

        self.default_mxid = self.get_mxid_from_id(pk)
        self.default_mxid_intent = self.az.intent.user(self.default_mxid)
        self.intent = self._fresh_intent()

    @classmethod
    def init_cls(cls, bridge: "InstagramBridge") -> AsyncIterable[Awaitable[None]]:
        cls.config = bridge.config
        cls.loop = bridge.loop
        cls.mx = bridge.matrix
        cls.az = bridge.az
        cls.hs_domain = cls.config["homeserver.domain"]
        cls.mxid_template = SimpleTemplate(
            cls.config["bridge.username_template"],
            "userid",
            prefix="@",
            suffix=f":{cls.hs_domain}",
            type=int,
        )
        cls.sync_with_custom_puppets = cls.config["bridge.sync_with_custom_puppets"]
        cls.homeserver_url_map = {
            server: URL(url)
            for server, url in cls.config["bridge.double_puppet_server_map"].items()
        }
        cls.allow_discover_url = cls.config["bridge.double_puppet_allow_discovery"]
        cls.login_shared_secret_map = {
            server: secret.encode("utf-8")
            for server, secret in cls.config["bridge.login_shared_secret_map"].items()
        }
        cls.login_device_name = "Instagram Bridge"
        return (puppet.try_start() async for puppet in cls.all_with_custom_mxid())

    @property
    def igpk(self) -> int:
        return self.pk

    def intent_for(self, portal: p.Portal) -> IntentAPI:
        if portal.other_user_pk == self.pk:
            return self.default_mxid_intent
        return self.intent

    def need_backfill_invite(self, portal: p.Portal) -> bool:
        return (
            portal.other_user_pk != self.pk
            and (self.is_real_user or portal.is_direct)
            and self.config["bridge.backfill.invite_own_puppet"]
        )

    async def update_info(self, info: BaseResponseUser, source: u.User) -> None:
        update = False
        update = await self._update_name(info) or update
        update = await self._update_avatar(info, source) or update
        if update:
            await self.update()

    @classmethod
    def _get_displayname(cls, info: BaseResponseUser) -> str:
        return cls.config["bridge.displayname_template"].format(
            displayname=info.full_name or info.username, id=info.pk, username=info.username
        )

    async def _update_name(self, info: BaseResponseUser) -> bool:
        name = self._get_displayname(info)
        if name != self.name:
            self.name = name
            try:
                await self.default_mxid_intent.set_displayname(self.name)
                self.name_set = True
            except Exception:
                self.log.exception("Failed to update displayname")
                self.name_set = False
            return True
        return False

    async def _update_avatar(self, info: BaseResponseUser, source: u.User) -> bool:
        pic_id = (
            f"id_{info.profile_pic_id}.jpg"
            if info.profile_pic_id
            else os.path.basename(URL(info.profile_pic_url).path)
        )
        if pic_id != self.photo_id or not self.avatar_set:
            self.photo_id = pic_id
            if info.has_anonymous_profile_picture:
                mxc = ""
            else:
                async with source.client.raw_http_get(info.profile_pic_url) as resp:
                    content_type = resp.headers["Content-Type"]
                    resp_data = await resp.read()
                mxc = await self.default_mxid_intent.upload_media(
                    data=resp_data,
                    mime_type=content_type,
                    filename=pic_id,
                    async_upload=self.config["homeserver.async_media"],
                )
            try:
                await self.default_mxid_intent.set_avatar_url(mxc)
                self.avatar_set = True
                self.photo_mxc = mxc
            except Exception:
                self.log.exception("Failed to update avatar")
                self.avatar_set = False
            return True
        return False

    async def default_puppet_should_leave_room(self, room_id: RoomID) -> bool:
        portal = await p.Portal.get_by_mxid(room_id)
        return portal and portal.other_user_pk != self.pk

    # region Database getters

    def _add_to_cache(self) -> None:
        self.by_pk[self.pk] = self
        if self.custom_mxid:
            self.by_custom_mxid[self.custom_mxid] = self

    async def save(self) -> None:
        await self.update()

    @classmethod
    async def get_by_mxid(cls, mxid: UserID, create: bool = True) -> Puppet | None:
        pk = cls.get_id_from_mxid(mxid)
        if pk:
            return await cls.get_by_pk(pk, create=create)
        return None

    @classmethod
    @async_getter_lock
    async def get_by_custom_mxid(cls, mxid: UserID) -> Puppet | None:
        try:
            return cls.by_custom_mxid[mxid]
        except KeyError:
            pass

        puppet = cast(cls, await super().get_by_custom_mxid(mxid))
        if puppet:
            puppet._add_to_cache()
            return puppet

        return None

    @classmethod
    def get_id_from_mxid(cls, mxid: UserID) -> int | None:
        return cls.mxid_template.parse(mxid)

    @classmethod
    def get_mxid_from_id(cls, pk: int) -> UserID:
        return UserID(cls.mxid_template.format_full(pk))

    @classmethod
    @async_getter_lock
    async def get_by_pk(cls, pk: int, *, create: bool = True) -> Puppet | None:
        try:
            return cls.by_pk[pk]
        except KeyError:
            pass

        puppet = cast(cls, await super().get_by_pk(pk))
        if puppet is not None:
            puppet._add_to_cache()
            return puppet

        if create:
            puppet = cls(pk)
            await puppet.insert()
            puppet._add_to_cache()
            return puppet

        return None

    @classmethod
    async def all_with_custom_mxid(cls) -> AsyncGenerator[Puppet, None]:
        puppets = await super().all_with_custom_mxid()
        puppet: cls
        for index, puppet in enumerate(puppets):
            try:
                yield cls.by_pk[puppet.pk]
            except KeyError:
                puppet._add_to_cache()
                yield puppet

    # endregion



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from attr import dataclass


@dataclass
class Connect:
    pass


@dataclass
class Disconnect:
    reason: str


@dataclass
class NewSequenceID:
    seq_id: int
    snapshot_at_ms: int



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional

from attr import dataclass

from mautrix.types import SerializableAttrs, field

from .login import LoginResponseUser


@dataclass
class ChallengeStateData(SerializableAttrs):
    # Only for reset step
    choice: Optional[str] = None
    fb_access_token: Optional[str] = None
    big_blue_token: Optional[str] = None
    google_oauth_token: Optional[str] = None
    email: Optional[str] = None

    # Only for verify email step
    security_code: Optional[str] = None
    resend_delay: Optional[int] = None
    contact_point: Optional[str] = None
    form_type: Optional[str] = None


@dataclass
class ChallengeContext(SerializableAttrs):
    step_name: Optional[str] = None
    challenge_type_enum: Optional[str] = None
    cni: Optional[int] = None
    is_stateless: bool = False
    present_as_modal: bool = False


@dataclass(kw_only=True)
class ChallengeStateResponse(SerializableAttrs):
    # TODO enum?
    step_name: Optional[str] = None
    step_data: Optional[ChallengeStateData] = None
    logged_in_user: Optional[LoginResponseUser] = None
    user_id: Optional[int] = None
    nonce_code: Optional[str] = None
    # TODO enum?
    action: Optional[str] = None
    status: str

    flow_render_type: Optional[int] = None
    bloks_action: Optional[str] = None
    challenge_context_str: Optional[str] = field(default=None, json="challenge_context")
    challenge_type_enum_str: Optional[str] = None

    @property
    def challenge_context(self) -> ChallengeContext:
        return ChallengeContext.parse_json(self.challenge_context_str)



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from asyncpg import Connection

from mautrix.util.async_db import UpgradeTable

upgrade_table = UpgradeTable()


@upgrade_table.register(description="Initial revision")
async def upgrade_v1(conn: Connection) -> None:
    await conn.execute(
        """CREATE TABLE portal (
        thread_id     TEXT,
        receiver      BIGINT,
        other_user_pk BIGINT,
        mxid          TEXT,
        name          TEXT,
        encrypted     BOOLEAN NOT NULL DEFAULT false,
        PRIMARY KEY (thread_id, receiver)
    )"""
    )
    await conn.execute(
        """CREATE TABLE "user" (
        mxid        TEXT PRIMARY KEY,
        igpk        BIGINT,
        state       jsonb,
        notice_room TEXT
    )"""
    )
    await conn.execute(
        """CREATE TABLE puppet (
        pk            BIGINT PRIMARY KEY,
        name          TEXT,
        username      TEXT,
        photo_id      TEXT,
        photo_mxc     TEXT,
        name_set      BOOLEAN NOT NULL DEFAULT false,
        avatar_set    BOOLEAN NOT NULL DEFAULT false,
        is_registered BOOLEAN NOT NULL DEFAULT false,
        custom_mxid   TEXT,
        access_token  TEXT,
        next_batch    TEXT,
        base_url      TEXT
    )"""
    )
    await conn.execute(
        """CREATE TABLE user_portal (
        "user"          BIGINT,
        portal          TEXT,
        portal_receiver BIGINT,
        in_community    BOOLEAN NOT NULL DEFAULT false,
        FOREIGN KEY (portal, portal_receiver) REFERENCES portal(thread_id, receiver)
            ON UPDATE CASCADE ON DELETE CASCADE
    )"""
    )
    await conn.execute(
        """CREATE TABLE message (
        mxid     TEXT NOT NULL,
        mx_room  TEXT NOT NULL,
        item_id  TEXT,
        receiver BIGINT,
        sender   BIGINT NOT NULL,
        PRIMARY KEY (item_id, receiver),
        UNIQUE (mxid, mx_room)
    )"""
    )
    await conn.execute(
        """CREATE TABLE reaction (
        mxid        TEXT NOT NULL,
        mx_room     TEXT NOT NULL,
        ig_item_id  TEXT,
        ig_receiver BIGINT,
        ig_sender   BIGINT,
        reaction    TEXT NOT NULL,
        PRIMARY KEY (ig_item_id, ig_receiver, ig_sender),
        FOREIGN KEY (ig_item_id, ig_receiver) REFERENCES message(item_id, receiver)
            ON DELETE CASCADE ON UPDATE CASCADE,
        UNIQUE (mxid, mx_room)
    )"""
    )


@upgrade_table.register(description="Add name_set and avatar_set to portal table")
async def upgrade_v2(conn: Connection) -> None:
    await conn.execute("ALTER TABLE portal ADD COLUMN avatar_url TEXT")
    await conn.execute("ALTER TABLE portal ADD COLUMN name_set BOOLEAN NOT NULL DEFAULT false")
    await conn.execute("ALTER TABLE portal ADD COLUMN avatar_set BOOLEAN NOT NULL DEFAULT false")
    await conn.execute("UPDATE portal SET name_set=true WHERE name<>''")


@upgrade_table.register(description="Add relay user field to portal table")
async def upgrade_v3(conn: Connection) -> None:
    await conn.execute("ALTER TABLE portal ADD COLUMN relay_user_id TEXT")


@upgrade_table.register(description="Add client context field to message table")
async def upgrade_v4(conn: Connection) -> None:
    await conn.execute("ALTER TABLE message ADD COLUMN client_context TEXT")


@upgrade_table.register(description="Add ig_timestamp field to message table")
async def upgrade_v5(conn: Connection) -> None:
    await conn.execute("ALTER TABLE message ADD COLUMN ig_timestamp BIGINT")


@upgrade_table.register(description="Allow hidden events in message table")
async def upgrade_v6(conn: Connection) -> None:
    await conn.execute("ALTER TABLE message ALTER COLUMN mxid DROP NOT NULL")


@upgrade_table.register(description="Store reaction timestamps")
async def upgrade_v7(conn: Connection) -> None:
    await conn.execute("ALTER TABLE reaction ADD COLUMN mx_timestamp BIGINT")


@upgrade_table.register(description="Store sync sequence ID in user table")
async def upgrade_v8(conn: Connection) -> None:
    await conn.execute('ALTER TABLE "user" ADD COLUMN seq_id BIGINT')
    await conn.execute('ALTER TABLE "user" ADD COLUMN snapshot_at_ms BIGINT')



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, List, Optional

from attr import dataclass

from mautrix.types import SerializableAttrs

from .thread import Thread, ThreadUser


@dataclass
class DMInboxCursor(SerializableAttrs):
    cursor_timestamp_seconds: int
    cursor_thread_v2_id: int


@dataclass
class DMInbox(SerializableAttrs):
    threads: List[Thread]
    has_older: bool
    unseen_count: int
    unseen_count_ts: int
    blended_inbox_enabled: bool
    prev_cursor: Optional[DMInboxCursor] = None
    next_cursor: Optional[DMInboxCursor] = None
    newest_cursor: Optional[str] = None
    oldest_cursor: Optional[str] = None


@dataclass
class DMInboxResponse(SerializableAttrs):
    status: str
    seq_id: int
    snapshot_at_ms: int
    pending_requests_total: int
    has_pending_top_requests: bool
    viewer: ThreadUser
    inbox: DMInbox
    # TODO type
    most_recent_inviter: Any = None


@dataclass
class DMThreadResponse(SerializableAttrs):
    thread: Thread
    status: str



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Optional

from attr import dataclass

from mautrix.types import SerializableAttrs


@dataclass
class SpamResponse(SerializableAttrs):
    feedback_title: str
    feedback_message: str
    feedback_url: str
    feedback_appeal_label: str
    feedback_ignore_label: str
    feedback_action: str
    message: str = "feedback_required"
    spam: bool = True
    status: str = "fail"
    error_type: Optional[str] = None


@dataclass
class ChallengeData(SerializableAttrs):
    url: str
    api_path: str
    hide_webview_header: bool
    lock: bool
    logout: bool
    native_flow: bool


@dataclass
class ChallengeResponse(SerializableAttrs):
    message: str  # challenge_required
    status: str  # fail
    challenge: ChallengeData
    error_type: Optional[str] = None


@dataclass
class CheckpointResponse(SerializableAttrs):
    message: str  # checkpoint_required
    status: str  # fail
    checkpoint_url: Optional[str] = None
    lock: bool = False
    flow_render_type: int = 0


@dataclass
class ConsentData(SerializableAttrs):
    headline: str
    content: str
    button_text: str


@dataclass
class ConsentRequiredResponse(SerializableAttrs):
    message: str  # consent_required
    status: str  # fail
    consent_data: ConsentData


@dataclass
class LoginRequiredResponse(SerializableAttrs):
    # TODO enum?
    logout_reason: int
    message: str  # login_required or user_has_logged_out
    status: str  # fail
    error_title: Optional[str] = None
    error_body: Optional[str] = None


@dataclass
class LoginErrorResponseButton(SerializableAttrs):
    title: str
    action: str


@dataclass
class LoginPhoneVerificationSettings(SerializableAttrs):
    max_sms_count: int
    resend_sms_delay_sec: int
    robocall_count_down_time_sec: int
    robocall_after_max_sms: bool


@dataclass
class LoginTwoFactorInfo(SerializableAttrs):
    username: str
    sms_two_factor_on: bool
    totp_two_factor_on: bool
    obfuscated_phone_number: str
    two_factor_identifier: str
    show_messenger_code_option: bool
    show_new_login_screen: bool
    show_trusted_device_option: bool
    should_opt_in_trusted_device_option: bool
    pending_trusted_notification: bool
    # TODO type
    # sms_not_allowed_reason: Any
    phone_verification_settings: Optional[LoginPhoneVerificationSettings] = None


@dataclass
class LoginErrorResponse(SerializableAttrs):
    message: str
    status: str
    error_type: str
    error_title: Optional[str] = None
    buttons: Optional[List[LoginErrorResponseButton]] = None
    invalid_credentials: Optional[bool] = None
    two_factor_required: Optional[bool] = None
    two_factor_info: Optional[LoginTwoFactorInfo] = None
    phone_verification_settings: Optional[LoginPhoneVerificationSettings] = None

    # FB login error fields
    account_created: Optional[bool] = None
    dryrun_passed: Optional[bool] = None



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, AsyncIterable, Awaitable, cast
import asyncio
import logging
import time

from mauigpapi import AndroidAPI, AndroidMQTT, AndroidState
from mauigpapi.errors import (
    IGChallengeError,
    IGCheckpointError,
    IGConsentRequiredError,
    IGNotLoggedInError,
    IGRateLimitError,
    IGUserIDNotFoundError,
    IrisSubscribeError,
    MQTTConnectionUnauthorized,
    MQTTNotConnected,
    MQTTNotLoggedIn,
)
from mauigpapi.mqtt import (
    Connect,
    Disconnect,
    GraphQLSubscription,
    NewSequenceID,
    SkywalkerSubscription,
)
from mauigpapi.types import (
    ActivityIndicatorData,
    CurrentUser,
    MessageSyncEvent,
    Operation,
    RealtimeDirectEvent,
    Thread,
    ThreadSyncEvent,
    TypingStatus,
)
from mautrix.appservice import AppService
from mautrix.bridge import BaseUser, async_getter_lock
from mautrix.types import EventID, MessageType, RoomID, TextMessageEventContent, UserID
from mautrix.util.bridge_state import BridgeState, BridgeStateEvent
from mautrix.util.logging import TraceLogger
from mautrix.util.opt_prometheus import Gauge, Summary, async_time

from . import portal as po, puppet as pu
from .config import Config
from .db import Portal as DBPortal, User as DBUser

if TYPE_CHECKING:
    from .__main__ import InstagramBridge

METRIC_MESSAGE = Summary("bridge_on_message", "calls to handle_message")
METRIC_THREAD_SYNC = Summary("bridge_on_thread_sync", "calls to handle_thread_sync")
METRIC_RTD = Summary("bridge_on_rtd", "calls to handle_rtd")
METRIC_LOGGED_IN = Gauge("bridge_logged_in", "Users logged into the bridge")
METRIC_CONNECTED = Gauge("bridge_connected", "Bridged users connected to Instagram")

BridgeState.human_readable_errors.update(
    {
        "ig-connection-error": "Instagram disconnected unexpectedly",
        "ig-refresh-connection-error": "Reconnecting failed again after refresh: {message}",
        "ig-connection-fatal-error": "Instagram disconnected unexpectedly",
        "ig-auth-error": "Authentication error from Instagram: {message}",
        "ig-checkpoint": "Instagram checkpoint error. Please check the Instagram website.",
        "ig-consent-required": "Instagram requires a consent update. Please check the Instagram website.",
        "ig-checkpoint-locked": "Instagram checkpoint error. Please check the Instagram website.",
        "ig-rate-limit": "Got Instagram ratelimit error, waiting a few minutes before retrying...",
        "ig-disconnected": None,
        "ig-no-mqtt": "You're not connected to Instagram",
        "logged-out": "You're not logged into Instagram",
    }
)


class User(DBUser, BaseUser):
    ig_base_log: TraceLogger = logging.getLogger("mau.instagram")
    _activity_indicator_ids: dict[str, int] = {}
    by_mxid: dict[UserID, User] = {}
    by_igpk: dict[int, User] = {}
    config: Config
    az: AppService
    loop: asyncio.AbstractEventLoop

    client: AndroidAPI | None
    mqtt: AndroidMQTT | None
    _listen_task: asyncio.Task | None = None
    _seq_id_save_task: asyncio.Task | None

    permission_level: str
    username: str | None

    _notice_room_lock: asyncio.Lock
    _notice_send_lock: asyncio.Lock
    _is_logged_in: bool
    _is_connected: bool
    shutdown: bool
    remote_typing_status: TypingStatus | None

    def __init__(
        self,
        mxid: UserID,
        igpk: int | None = None,
        state: AndroidState | None = None,
        notice_room: RoomID | None = None,
        seq_id: int | None = None,
        snapshot_at_ms: int | None = None,
    ) -> None:
        super().__init__(
            mxid=mxid,
            igpk=igpk,
            state=state,
            notice_room=notice_room,
            seq_id=seq_id,
            snapshot_at_ms=snapshot_at_ms,
        )
        BaseUser.__init__(self)
        self._notice_room_lock = asyncio.Lock()
        self._notice_send_lock = asyncio.Lock()
        perms = self.config.get_permissions(mxid)
        self.relay_whitelisted, self.is_whitelisted, self.is_admin, self.permission_level = perms
        self.client = None
        self.mqtt = None
        self.username = None
        self._is_logged_in = False
        self._is_connected = False
        self._is_refreshing = False
        self.shutdown = False
        self._listen_task = None
        self.remote_typing_status = None
        self._seq_id_save_task = None

    @classmethod
    def init_cls(cls, bridge: "InstagramBridge") -> AsyncIterable[Awaitable[None]]:
        cls.bridge = bridge
        cls.config = bridge.config
        cls.az = bridge.az
        cls.loop = bridge.loop
        return (user.try_connect() async for user in cls.all_logged_in())

    # region Connection management

    async def is_logged_in(self) -> bool:
        return bool(self.client) and self._is_logged_in

    async def get_puppet(self) -> pu.Puppet | None:
        if not self.igpk:
            return None
        return await pu.Puppet.get_by_pk(self.igpk)

    async def get_portal_with(self, puppet: pu.Puppet, create: bool = True) -> po.Portal | None:
        if not self.igpk:
            return None
        portal = await po.Portal.find_private_chat(self.igpk, puppet.pk)
        if portal:
            return portal
        if create:
            # TODO add error handling somewhere
            thread = await self.client.create_group_thread([puppet.pk])
            portal = await po.Portal.get_by_thread(thread, self.igpk)
            await portal.update_info(thread, self)
            return portal
        return None

    async def try_connect(self) -> None:
        try:
            await self.connect()
        except Exception as e:
            self.log.exception("Error while connecting to Instagram")
            await self.push_bridge_state(
                BridgeStateEvent.UNKNOWN_ERROR, info={"python_error": str(e)}
            )

    @property
    def api_log(self) -> TraceLogger:
        return self.ig_base_log.getChild("http").getChild(self.mxid)

    @property
    def is_connected(self) -> bool:
        return bool(self.client) and bool(self.mqtt) and self._is_connected

    async def connect(self, user: CurrentUser | None = None) -> None:
        if not self.state:
            await self.push_bridge_state(BridgeStateEvent.BAD_CREDENTIALS, error="logged-out")
            return
        client = AndroidAPI(self.state, log=self.api_log)

        if not user:
            try:
                resp = await client.current_user()
                user = resp.user
            except IGNotLoggedInError as e:
                self.log.warning(f"Failed to connect to Instagram: {e}, logging out")
                await self.logout(error=e)
                return
            except IGCheckpointError as e:
                self.log.debug("Checkpoint error content: %s", e.body)
                raise
            except (IGChallengeError, IGConsentRequiredError) as e:
                await self._handle_checkpoint(e, on="connect", client=client)
                return
        self.client = client
        self._is_logged_in = True
        self.igpk = user.pk
        self.username = user.username
        await self.push_bridge_state(BridgeStateEvent.CONNECTING)
        self._track_metric(METRIC_LOGGED_IN, True)
        self.by_igpk[self.igpk] = self

        self.mqtt = AndroidMQTT(
            self.state, loop=self.loop, log=self.ig_base_log.getChild("mqtt").getChild(self.mxid)
        )
        self.mqtt.add_event_handler(Connect, self.on_connect)
        self.mqtt.add_event_handler(Disconnect, self.on_disconnect)
        self.mqtt.add_event_handler(NewSequenceID, self.update_seq_id)
        self.mqtt.add_event_handler(MessageSyncEvent, self.handle_message)
        self.mqtt.add_event_handler(ThreadSyncEvent, self.handle_thread_sync)
        self.mqtt.add_event_handler(RealtimeDirectEvent, self.handle_rtd)

        await self.update()

        self.loop.create_task(self._try_sync_puppet(user))
        if not self.seq_id or self.config["bridge.resync_on_startup"]:
            self.loop.create_task(self._try_sync())
        else:
            self.log.debug("Connecting to MQTT directly as resync_on_startup is false")
            self.start_listen()

    async def on_connect(self, evt: Connect) -> None:
        self.log.debug("Connected to Instagram")
        self._track_metric(METRIC_CONNECTED, True)
        self._is_connected = True
        await self.send_bridge_notice("Connected to Instagram")
        await self.push_bridge_state(BridgeStateEvent.CONNECTED)

    async def on_disconnect(self, evt: Disconnect) -> None:
        self.log.debug("Disconnected from Instagram")
        self._track_metric(METRIC_CONNECTED, False)
        self._is_connected = False

    # TODO this stuff could probably be moved to mautrix-python
    async def get_notice_room(self) -> RoomID:
        if not self.notice_room:
            async with self._notice_room_lock:
                # If someone already created the room while this call was waiting,
                # don't make a new room
                if self.notice_room:
                    return self.notice_room
                creation_content = {}
                if not self.config["bridge.federate_rooms"]:
                    creation_content["m.federate"] = False
                self.notice_room = await self.az.intent.create_room(
                    is_direct=True,
                    invitees=[self.mxid],
                    topic="Instagram bridge notices",
                    creation_content=creation_content,
                )
                await self.update()
        return self.notice_room

    async def fill_bridge_state(self, state: BridgeState) -> None:
        await super().fill_bridge_state(state)
        if not state.remote_id:
            if self.igpk:
                state.remote_id = str(self.igpk)
            else:
                try:
                    state.remote_id = self.state.user_id
                except IGUserIDNotFoundError:
                    state.remote_id = None
        if self.username:
            state.remote_name = f"@{self.username}"

    async def get_bridge_states(self) -> list[BridgeState]:
        if not self.state:
            return []
        state = BridgeState(state_event=BridgeStateEvent.UNKNOWN_ERROR)
        if self.is_connected:
            state.state_event = BridgeStateEvent.CONNECTED
        elif self._is_refreshing or self.mqtt:
            state.state_event = BridgeStateEvent.TRANSIENT_DISCONNECT
        return [state]

    async def send_bridge_notice(
        self,
        text: str,
        edit: EventID | None = None,
        state_event: BridgeStateEvent | None = None,
        important: bool = False,
        error_code: str | None = None,
        error_message: str | None = None,
        info: dict | None = None,
    ) -> EventID | None:
        if state_event:
            await self.push_bridge_state(
                state_event,
                error=error_code,
                message=error_message if error_code else text,
                info=info,
            )
        if self.config["bridge.disable_bridge_notices"]:
            return None
        if not important and not self.config["bridge.unimportant_bridge_notices"]:
            self.log.debug("Not sending unimportant bridge notice: %s", text)
            return None
        event_id = None
        try:
            self.log.debug("Sending bridge notice: %s", text)
            content = TextMessageEventContent(
                body=text, msgtype=(MessageType.TEXT if important else MessageType.NOTICE)
            )
            if edit:
                content.set_edit(edit)
            # This is locked to prevent notices going out in the wrong order
            async with self._notice_send_lock:
                event_id = await self.az.intent.send_message(await self.get_notice_room(), content)
        except Exception:
            self.log.warning("Failed to send bridge notice", exc_info=True)
        return edit or event_id

    async def _try_sync_puppet(self, user_info: CurrentUser) -> None:
        puppet = await pu.Puppet.get_by_pk(self.igpk)
        try:
            await puppet.update_info(user_info, self)
        except Exception:
            self.log.exception("Failed to update own puppet info")
        try:
            if puppet.custom_mxid != self.mxid and puppet.can_auto_login(self.mxid):
                self.log.info(f"Automatically enabling custom puppet")
                await puppet.switch_mxid(access_token="auto", mxid=self.mxid)
        except Exception:
            self.log.exception("Failed to automatically enable custom puppet")

    async def _try_sync(self) -> None:
        try:
            await self.sync()
        except Exception as e:
            self.log.exception("Exception while syncing")
            if isinstance(e, IGCheckpointError):
                self.log.debug("Checkpoint error content: %s", e.body)
            await self.push_bridge_state(
                BridgeStateEvent.UNKNOWN_ERROR, info={"python_error": str(e)}
            )

    async def get_direct_chats(self) -> dict[UserID, list[RoomID]]:
        return {
            pu.Puppet.get_mxid_from_id(portal.other_user_pk): [portal.mxid]
            for portal in await DBPortal.find_private_chats_of(self.igpk)
            if portal.mxid
        }

    async def refresh(self, resync: bool = True) -> None:
        self._is_refreshing = True
        try:
            await self.stop_listen()
            if resync:
                retry_count = 0
                minutes = 1
                while True:
                    try:
                        await self.sync()
                        return
                    except Exception as e:
                        if retry_count >= 4 and minutes < 10:
                            minutes += 1
                        retry_count += 1
                        s = "s" if minutes != 1 else ""
                        self.log.exception(
                            f"Error while syncing for refresh, retrying in {minutes} minute{s}"
                        )
                        if isinstance(e, IGCheckpointError):
                            self.log.debug("Checkpoint error content: %s", e.body)
                        await self.push_bridge_state(
                            BridgeStateEvent.UNKNOWN_ERROR,
                            error="unknown-error",
                            message="An unknown error occurred while connecting to Instagram",
                            info={"python_error": str(e)},
                        )
                        await asyncio.sleep(minutes * 60)
            else:
                self.start_listen()
        finally:
            self._is_refreshing = False

    async def _handle_checkpoint(
        self,
        e: IGChallengeError | IGConsentRequiredError,
        on: str,
        client: AndroidAPI | None = None,
    ) -> None:
        self.log.warning(f"Got checkpoint error on {on}: {e.body.serialize()}")
        client = client or self.client
        self.client = None
        self.mqtt = None
        if isinstance(e, IGConsentRequiredError):
            await self.push_bridge_state(
                BridgeStateEvent.BAD_CREDENTIALS,
                error="ig-consent-required",
                info=e.body.serialize(),
            )
            return
        error_code = "ig-checkpoint"
        try:
            resp = await client.challenge_reset()
            info = {
                "challenge_context": (
                    resp.challenge_context.serialize() if resp.challenge_context_str else None
                ),
                "step_name": resp.step_name,
                "step_data": resp.step_data.serialize() if resp.step_data else None,
                "user_id": resp.user_id,
                "action": resp.action,
                "status": resp.status,
                "challenge": e.body.challenge.serialize() if e.body.challenge else None,
            }
            self.log.debug(f"Challenge state: {resp.serialize()}")
            if resp.challenge_context.challenge_type_enum == "HACKED_LOCK":
                error_code = "ig-checkpoint-locked"
        except Exception:
            self.log.exception("Error resetting challenge state")
            info = {"challenge": e.body.challenge.serialize() if e.body.challenge else None}
        await self.push_bridge_state(BridgeStateEvent.BAD_CREDENTIALS, error=error_code, info=info)

    async def _sync_thread(self, thread: Thread, allow_create: bool) -> None:
        portal = await po.Portal.get_by_thread(thread, self.igpk)
        if portal.mxid:
            self.log.debug(f"{thread.thread_id} has a portal, syncing and backfilling...")
            await portal.update_matrix_room(self, thread, backfill=True)
        elif allow_create:
            self.log.debug(f"{thread.thread_id} has been active recently, creating portal...")
            await portal.create_matrix_room(self, thread)
        else:
            self.log.debug(f"{thread.thread_id} is not active and doesn't have a portal")

    async def sync(self) -> None:
        sleep_minutes = 2
        while True:
            try:
                resp = await self.client.get_inbox()
                break
            except IGNotLoggedInError as e:
                self.log.exception("Got not logged in error while syncing")
                await self.logout(error=e)
                return
            except IGRateLimitError as e:
                self.log.error(
                    "Got ratelimit error while trying to get inbox (%s), retrying in %d minutes",
                    e.body,
                    sleep_minutes,
                )
                await self.push_bridge_state(
                    BridgeStateEvent.TRANSIENT_DISCONNECT, error="ig-rate-limit"
                )
                await asyncio.sleep(sleep_minutes * 60)
                sleep_minutes += 2
            except IGCheckpointError as e:
                self.log.debug("Checkpoint error content: %s", e.body)
                raise
            except (IGChallengeError, IGConsentRequiredError) as e:
                await self._handle_checkpoint(e, on="sync")
                return

        self.seq_id = resp.seq_id
        self.snapshot_at_ms = resp.snapshot_at_ms
        await self.save_seq_id()

        if not self._listen_task:
            self.start_listen(is_after_sync=True)

        max_age = self.config["bridge.portal_create_max_age"] * 1_000_000
        limit = self.config["bridge.chat_sync_limit"]
        create_limit = self.config["bridge.chat_create_limit"]
        min_active_at = (time.time() * 1_000_000) - max_age
        i = 0
        await self.push_bridge_state(BridgeStateEvent.BACKFILLING)
        async for thread in self.client.iter_inbox(start_at=resp):
            try:
                await self._sync_thread(
                    thread=thread,
                    allow_create=thread.last_activity_at > min_active_at and i < create_limit,
                )
            except Exception:
                self.log.exception(f"Error syncing thread {thread.thread_id}")
            i += 1
            if i >= limit:
                break
        try:
            await self.update_direct_chats()
        except Exception:
            self.log.exception("Error updating direct chat list")

    def start_listen(self, is_after_sync: bool = False) -> None:
        self.shutdown = False
        task = self._listen(
            seq_id=self.seq_id, snapshot_at_ms=self.snapshot_at_ms, is_after_sync=is_after_sync
        )
        self._listen_task = self.loop.create_task(task)

    async def fetch_user_and_reconnect(self) -> None:
        self.log.debug("Refetching current user after disconnection")
        try:
            resp = await self.client.current_user()
        except IGNotLoggedInError as e:
            self.log.warning(f"Failed to reconnect to Instagram: {e}, logging out")
            await self.logout(error=e)
            return
        except (IGChallengeError, IGConsentRequiredError) as e:
            await self._handle_checkpoint(e, on="reconnect")
            return
        except Exception as e:
            self.log.exception("Error while reconnecting to Instagram")
            if isinstance(e, IGCheckpointError):
                self.log.debug("Checkpoint error content: %s", e.body)
            await self.push_bridge_state(
                BridgeStateEvent.UNKNOWN_ERROR, info={"python_error": str(e)}
            )
            return
        else:
            self.log.debug(f"Confirmed current user {resp.user.pk}")
            self.start_listen()

    async def _listen(self, seq_id: int, snapshot_at_ms: int, is_after_sync: bool) -> None:
        try:
            await self.mqtt.listen(
                graphql_subs={
                    GraphQLSubscription.app_presence(),
                    GraphQLSubscription.direct_typing(self.state.user_id),
                    GraphQLSubscription.direct_status(),
                },
                skywalker_subs={
                    SkywalkerSubscription.direct_sub(self.state.user_id),
                    SkywalkerSubscription.live_sub(self.state.user_id),
                },
                seq_id=seq_id,
                snapshot_at_ms=snapshot_at_ms,
            )
        except IrisSubscribeError as e:
            if is_after_sync:
                self.log.exception("Got IrisSubscribeError right after refresh")
                await self.send_bridge_notice(
                    f"Reconnecting failed again after refresh: {e}",
                    important=True,
                    state_event=BridgeStateEvent.UNKNOWN_ERROR,
                    error_code="ig-refresh-connection-error",
                    error_message=str(e),
                    info={"python_error": str(e)},
                )
            else:
                self.log.warning(f"Got IrisSubscribeError {e}, refreshing...")
                asyncio.create_task(self.refresh())
        except (MQTTNotConnected, MQTTNotLoggedIn, MQTTConnectionUnauthorized) as e:
            self.log.warning(f"Unexpected connection error: {e}")
            await self.send_bridge_notice(
                f"Error in listener: {e}",
                important=True,
                state_event=BridgeStateEvent.UNKNOWN_ERROR,
                error_code="ig-connection-error",
            )
            self.mqtt.disconnect()
            asyncio.create_task(self.fetch_user_and_reconnect())
        except Exception as e:
            self.log.exception("Fatal error in listener")
            await self.send_bridge_notice(
                "Fatal error in listener (see logs for more info)",
                state_event=BridgeStateEvent.UNKNOWN_ERROR,
                important=True,
                error_code="ig-unknown-connection-error",
                info={"python_error": str(e)},
            )
            self.mqtt.disconnect()
        else:
            if not self.shutdown:
                await self.send_bridge_notice(
                    "Instagram connection closed without error",
                    state_event=BridgeStateEvent.UNKNOWN_ERROR,
                    error_code="ig-disconnected",
                )
        finally:
            self._listen_task = None
            self._is_connected = False
            self._track_metric(METRIC_CONNECTED, False)

    async def stop_listen(self) -> None:
        if self.mqtt:
            self.shutdown = True
            self.mqtt.disconnect()
            if self._listen_task:
                await self._listen_task
            self.shutdown = False
        self._track_metric(METRIC_CONNECTED, False)
        self._is_connected = False
        await self.update()

    async def logout(self, error: IGNotLoggedInError | None = None) -> None:
        if self.client and error is None:
            try:
                await self.client.logout(one_tap_app_login=False)
            except Exception:
                self.log.debug("Exception logging out", exc_info=True)
        if self.mqtt:
            self.mqtt.disconnect()
        self._track_metric(METRIC_CONNECTED, False)
        self._track_metric(METRIC_LOGGED_IN, False)
        if error is None:
            await self.push_bridge_state(BridgeStateEvent.LOGGED_OUT)
            puppet = await pu.Puppet.get_by_pk(self.igpk, create=False)
            if puppet and puppet.is_real_user:
                await puppet.switch_mxid(None, None)
            try:
                del self.by_igpk[self.igpk]
            except KeyError:
                pass
            self.igpk = None
        else:
            self.log.debug("Auth error body: %s", error.body.serialize())
            await self.send_bridge_notice(
                f"You have been logged out of Instagram: {error.proper_message}",
                important=True,
                state_event=BridgeStateEvent.BAD_CREDENTIALS,
                error_code="ig-auth-error",
                error_message=error.proper_message,
            )
        self.client = None
        self.mqtt = None
        self.state = None
        self.seq_id = None
        self.snapshot_at_ms = None
        self._is_logged_in = False
        await self.update()

    # endregion
    # region Event handlers

    async def _save_seq_id_after_sleep(self) -> None:
        await asyncio.sleep(120)
        self._seq_id_save_task = None
        self.log.trace("Saving sequence ID %d/%d", self.seq_id, self.snapshot_at_ms)
        try:
            await self.save_seq_id()
        except Exception:
            self.log.exception("Error saving sequence ID")

    async def update_seq_id(self, evt: NewSequenceID) -> None:
        self.seq_id = evt.seq_id
        self.snapshot_at_ms = evt.snapshot_at_ms
        if not self._seq_id_save_task or self._seq_id_save_task.done():
            self.log.trace("Starting seq id save task (%d/%d)", evt.seq_id, evt.snapshot_at_ms)
            self._seq_id_save_task = asyncio.create_task(self._save_seq_id_after_sleep())
        else:
            self.log.trace("Not starting seq id save task (%d/%d)", evt.seq_id, evt.snapshot_at_ms)

    @async_time(METRIC_MESSAGE)
    async def handle_message(self, evt: MessageSyncEvent) -> None:
        portal = await po.Portal.get_by_thread_id(evt.message.thread_id, receiver=self.igpk)
        if not portal or not portal.mxid:
            self.log.debug("Got message in thread with no portal, getting info...")
            resp = await self.client.get_thread(evt.message.thread_id)
            portal = await po.Portal.get_by_thread(resp.thread, self.igpk)
            self.log.debug("Got info for unknown portal, creating room")
            await portal.create_matrix_room(self, resp.thread)
            if not portal.mxid:
                self.log.warning(
                    "Room creation appears to have failed, "
                    f"dropping message in {evt.message.thread_id}"
                )
                return
        self.log.trace(f"Received message sync event {evt.message}")
        sender = await pu.Puppet.get_by_pk(evt.message.user_id) if evt.message.user_id else None
        if evt.message.op == Operation.ADD:
            if not sender:
                # I don't think we care about adds with no sender
                return
            await portal.handle_instagram_item(self, sender, evt.message)
        elif evt.message.op == Operation.REMOVE:
            # Removes don't have a sender, only the message sender can unsend messages anyway
            await portal.handle_instagram_remove(evt.message.item_id)
        elif evt.message.op == Operation.REPLACE:
            await portal.handle_instagram_update(evt.message)

    @async_time(METRIC_THREAD_SYNC)
    async def handle_thread_sync(self, evt: ThreadSyncEvent) -> None:
        self.log.trace("Thread sync event content: %s", evt)
        portal = await po.Portal.get_by_thread(evt, receiver=self.igpk)
        if portal.mxid:
            self.log.debug("Got thread sync event for %s with existing portal", portal.thread_id)
            await portal.update_matrix_room(self, evt)
        elif evt.is_group:
            self.log.debug(
                "Got thread sync event for group %s without existing portal, creating room",
                portal.thread_id,
            )
            await portal.create_matrix_room(self, evt)
        else:
            self.log.debug(
                "Got thread sync event for DM %s without existing portal, ignoring",
                portal.thread_id,
            )

    @async_time(METRIC_RTD)
    async def handle_rtd(self, evt: RealtimeDirectEvent) -> None:
        if not isinstance(evt.value, ActivityIndicatorData):
            return

        now = int(time.time() * 1000)
        date = evt.value.timestamp_ms
        expiry = date + evt.value.ttl
        if expiry < now:
            return

        if evt.activity_indicator_id in self._activity_indicator_ids:
            return
        # TODO clear expired items from this dict
        self._activity_indicator_ids[evt.activity_indicator_id] = expiry

        puppet = await pu.Puppet.get_by_pk(int(evt.value.sender_id))
        portal = await po.Portal.get_by_thread_id(evt.thread_id, receiver=self.igpk)
        if not puppet or not portal or not portal.mxid:
            return

        is_typing = evt.value.activity_status != TypingStatus.OFF
        if puppet.pk == self.igpk:
            self.remote_typing_status = TypingStatus.TEXT if is_typing else TypingStatus.OFF
        await puppet.intent_for(portal).set_typing(
            portal.mxid, is_typing=is_typing, timeout=evt.value.ttl
        )

    # endregion
    # region Database getters

    def _add_to_cache(self) -> None:
        self.by_mxid[self.mxid] = self
        if self.igpk:
            self.by_igpk[self.igpk] = self

    @classmethod
    @async_getter_lock
    async def get_by_mxid(cls, mxid: UserID, *, create: bool = True) -> User | None:
        # Never allow ghosts to be users
        if pu.Puppet.get_id_from_mxid(mxid):
            return None
        try:
            return cls.by_mxid[mxid]
        except KeyError:
            pass

        user = cast(cls, await super().get_by_mxid(mxid))
        if user is not None:
            user._add_to_cache()
            return user

        if create:
            user = cls(mxid)
            await user.insert()
            user._add_to_cache()
            return user

        return None

    @classmethod
    @async_getter_lock
    async def get_by_igpk(cls, igpk: int) -> User | None:
        try:
            return cls.by_igpk[igpk]
        except KeyError:
            pass

        user = cast(cls, await super().get_by_igpk(igpk))
        if user is not None:
            user._add_to_cache()
            return user

        return None

    @classmethod
    async def all_logged_in(cls) -> AsyncGenerator[User, None]:
        users = await super().all_logged_in()
        user: cls
        for index, user in enumerate(users):
            try:
                yield cls.by_mxid[user.mxid]
            except KeyError:
                user._add_to_cache()
                yield user

    # endregion



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from enum import IntEnum


class TType(IntEnum):
    STOP = 0
    TRUE = 1
    FALSE = 2
    BYTE = 3
    I16 = 4
    I32 = 5
    I64 = 6
    # DOUBLE = 7
    BINARY = 8
    STRING = 8
    LIST = 9
    SET = 10
    MAP = 11
    STRUCT = 12

    # internal
    BOOL = 0xA1
